import geo_parameters as gp


def test_docstring():
    assert gp.wave.Hs.is_in([gp.wave.Hs, gp.wave.Tp])
    assert not gp.wave.Hs.is_in([gp.wave.Dirp, gp.wave.Tp])
    assert not gp.wave.Hs.is_in([])
    assert not gp.wave.Hs.is_in(["hs"])
    assert gp.wave.Hs.is_in([gp.wave.Hs(), "hs"])



from . import wave, ocean, wind
import inspect
from .metaparameter import MetaParameter
from .wind import WindParameter
from .wave import WaveParameter
from .ocean import OceanParameter
from .grid import GridParameter

BaseParameters = [
    MetaParameter,
    WindParameter,
    WaveParameter,
    OceanParameter,
    GridParameter,
]


def list_of_parameters() -> list:
    """Get a list of all available geo-parameters"""
    parameter_list = []
    modules_to_inspect = [wave, ocean, wind]
    for module in modules_to_inspect:
        for __, obj in inspect.getmembers(module):
            if inspect.isclass(obj):
                if issubclass(obj, MetaParameter) and not obj in BaseParameters:
                    parameter_list.append(obj)
    return parameter_list


def dict_of_parameters(short: bool = False, alias: bool = False) -> dict:
    """Get a dict of all availabel geo-parameters.

    Key is either standard_name (default), alias (alias=True) or non standard short name (short=True)
    """
    if short:
        return {c.name: c for c in list_of_parameters()}
    return {c.standard_name(alias=alias): c for c in list_of_parameters()}


def shortget(key: str):
    """Find a geo-parameter based on a short string. Warning, these short strings follow no standard convention."""
    return dict_of_parameters(short=True).get(key)


def get(key: str):
    """Find a geo-parameter based on a string of the standard_name (or an alias)."""
    return dict_of_parameters().get(key) or dict_of_parameters(alias=True).get(key)


def create_parameter_dict(parameter_strings: list[str]):
    """Create a dict of geo-parameters based on a list of strings of standard_names"""
    metaparameter_dict = {}
    for param in parameter_strings:
        val = get(param)
        if val is not None:
            metaparameter_dict[param] = val
    return metaparameter_dict



import geo_parameters as gp


def test_docstring():
    assert gp.wave.Hs.find_me_in([gp.wave.Hs, gp.wave.Tp]) == [gp.wave.Hs]
    assert (
        gp.wave.Hs.find_me_in([gp.wave.Hs, gp.wave.Tp], return_first=True) == gp.wave.Hs
    )
    assert gp.wave.Hs.find_me_in([gp.wave.Dirp, gp.wave.Tp]) == []
    assert gp.wave.Hs.find_me_in([], return_first=True) is None
    assert gp.wave.Hs.find_me_in(["hs"]) == []
    hs = gp.wave.Hs("swh")
    assert gp.wave.Hs.find_me_in([hs, "hs"])[0] == hs



from abc import ABC, abstractmethod
import numpy as np
from typing import Union
from pint import Unit

from .relationships import RELATIONSHIPS
from typing import Iterable


class MetaParameter(ABC):
    _cf = True

    def __init__(self, name: str = ""):
        self.name = name or self.name

    @staticmethod
    @abstractmethod
    def _long_name() -> str:
        pass

    @staticmethod
    @abstractmethod
    def _standard_name() -> str:
        pass

    @staticmethod
    @abstractmethod
    def _unit() -> str:
        pass

    @classmethod
    def cf(cls) -> bool:
        return cls._cf

    @classmethod
    def long_name(cls) -> str:
        return cls._long_name

    @classmethod
    def standard_name(cls, strict: bool = False, alias: bool = False) -> str:
        names = np.atleast_1d(cls._standard_name)
        if cls.cf() or not strict:
            if alias:
                return str(names[-1])
            return str(names[0])

        return ""

    @classmethod
    def standard_aliases(cls, strict=False) -> list[str]:
        """Returns a list of all aliases, possible a list of only one name"""
        if cls.cf() or not strict:
            if isinstance(cls._standard_name, list):
                return cls._standard_name
            else:
                return [cls._standard_name]

        return [""]

    @classmethod
    def units(cls) -> Unit:
        return cls._unit

    @classmethod
    def cf(cls) -> str:
        return cls._cf

    @classmethod
    def meta_dict(cls, alias: bool = False) -> dict:
        return {
            "short_name": cls.name,
            "long_name": cls.long_name(),
            "standard_name": cls.standard_name(alias=alias),
            "units": str(cls.units()),
        }

    @classmethod
    def find_me_in_ds(
        cls, ds, return_first: bool = False
    ) -> Union[list[str], str, None]:
        """Takes an Xarray Dataset and returns list of all variable names that matches the parameter based on standard_name

        If return_first = True, then returns first name as a sting"""
        data_vars = list(ds.data_vars)
        coords = list(ds.coords)

        if not return_first:
            vars = []

        for var in data_vars:
            if hasattr(ds.get(var), "standard_name"):
                if ds.get(var).standard_name in cls.standard_aliases():
                    if return_first:
                        return var
                    else:
                        vars.append(var)

        for var in coords:
            if hasattr(ds.get(var), "standard_name"):
                if ds.get(var).standard_name in cls.standard_aliases():
                    if return_first:
                        return var
                    else:
                        vars.append(var)

        if return_first:
            return None
        else:
            return vars

    @classmethod
    def i_am(cls) -> str:
        """Finds what role this parameter has in a possible relationship to other parameters.
        Roles are:
        'x', 'y', 'mag', 'dir', 'opposite_dir'
        'period', 'frequency', 'angular_frequency'
        """
        if cls.my_family() is None:
            return None
        for key, value in cls.my_family().items():
            if cls.is_same(value):
                return key
        return None

    @classmethod
    def is_same(cls, gp) -> bool:
        """Compares name of parameter and standard_name and returns true if they match

        param = gp.wave.Hs
        param.is_same(gp.wave.Hs) [True]
        param.is_same(gp.wave.Hs()) [True]
        param.is_same(gp.wave.Hs('other_name')) [True]
        param.is_same(gp.wave.Hmax) [False]
        """
        try:
            gp = gp()
        except TypeError:
            pass

        type_match = type(cls()).__name__ == type(gp).__name__
        try:
            std_name_match = cls.standard_name() == gp.standard_name()
            return type_match and std_name_match
        except AttributeError:
            return False

    @classmethod
    def is_in(cls, parameters: Iterable) -> bool:
        """Checks if the parameter class is in (using .is_same()) in the list of provided parameters

        Examples:
        >> gp.wave.Hs.is_in([gp.wave.Hs, gp.wave.Tp]) -> True
        >> gp.wave.Hs.is_in([gp.wave.Dirp, gp.wave.Tp]) -> False
        >> gp.wave.Hs.is_in([]) -> False
        >> gp.wave.Hs.is_in(['hs']) -> False
        >> gp.wave.Hs.is_in([gp.wave.Hs(), 'hs']) -> True
        """

        for param in parameters:
            if cls.is_same(param):
                return True
        return False

    @classmethod
    def find_me_in(
        cls, parameters: Iterable, return_first: bool = False
    ) -> Union[list, None]:
        """Gets the parameters in list that matches the class (using .is_same())

        Examples:
        >> gp.wave.Hs.find_me_in([gp.wave.Hs, gp.wave.Tp]) -> gp.wave.Hs
        >> gp.wave.Hs.find_me_in([gp.wave.Dirp, gp.wave.Tp]) -> None
        >> gp.wave.Hs.find_me_in([]) -> None
        >> gp.wave.Hs.find_me_in(['hs']) -> None
        >> swh = gp.wave.Hs("swh")
        >> gp.wave.Hs.find_me_in([swh, 'hs']) -> swh
        """
        matching_params = []
        for param in parameters:
            if cls.is_same(param):
                if return_first:
                    return param
                else:
                    matching_params.append(param)

        if return_first:
            return None
        else:
            return matching_params

    @classmethod
    def dir_type(cls) -> str:
        for std_name in cls.standard_aliases():
            if "to_direction" in std_name:
                return "to"
            if "from_direction" in std_name:
                return "from"
        return None



import geo_parameters as gp


def test_trivial():
    assert gp.wave.Hs.is_same(gp.wave.Hs)


def test_initialized():
    assert gp.wave.Hs.is_same(gp.wave.Hs())
    assert gp.wave.Hs().is_same(gp.wave.Hs)
    assert gp.wave.Hs().is_same(gp.wave.Hs())


def test_initialized_with_name():
    assert gp.wave.Hs.is_same(gp.wave.Hs("hsig"))
    assert gp.wave.Hs("hsig").is_same(gp.wave.Hs("hsig"))
    assert gp.wave.Hs("swh").is_same(gp.wave.Hs("hsig"))
    assert gp.wave.Hs("hsig").is_same(gp.wave.Hs)


def test_not_same():
    assert not gp.wave.Hs.is_same(gp.wind.XWind)
    assert not gp.wave.Hs.is_same(gp.wave.Tp)
    assert not gp.wave.Hs.is_same("hs")



import geo_parameters as gp

import pytest


def test_decode():
    assert gp.decode("test") == ("test", None)
    assert gp.decode(gp.wave.Hs) == ("hs", gp.wave.Hs)
    ii = gp.wave.Hs()
    assert gp.decode(ii) == ("hs", ii)
    ii = gp.wave.Hs("hsig")
    assert gp.decode(ii) == ("hsig", ii)

    with pytest.raises(TypeError):
        gp.decode(1)


def test_decode_init():
    var, param = gp.decode(gp.wave.Hs)
    assert gp.is_gp_class(param)

    var, param = gp.decode(gp.wave.Hs, init=True)
    assert gp.is_gp_instance(param)



from geo_parameters.metaparameter import MetaParameter
from geo_parameters.ureg import ureg
from typing import Optional, Union
from .relationships import _get_family_dict, _verify_param_type


class OceanParameter(MetaParameter):
    @classmethod
    def my_family(
        cls, param_type: Optional[str] = None
    ) -> Union[dict[str, type["OceanParameter"]], type["OceanParameter"], None]:
        """Returns the dictonary containing the parameters where cls is in.
        Use .my_family('direction') to get the parameter isntead of a dict"""

        _verify_param_type(param_type)
        family_dict = _get_family_dict(cls)

        if param_type is None:  # Return entire family_dict
            return_dict = {}
            for key, value in family_dict.items():
                # E.g. eval("WaterDepth"), which can't be done outside of this module
                return_dict[key] = eval(value)
            return return_dict
        else:  # Retrun class for requested parameter type
            return eval(family_dict.get(param_type, "None"))


class WaterDepth(OceanParameter):
    name = "depth"
    _long_name = "water_depth"
    _standard_name = "sea_floor_depth_below_sea_surface"
    _unit = ureg.m


class SeaLevel(OceanParameter):
    name = "eta"
    _long_name = "sea_surface_height"
    _standard_name = [
        "sea_surface_elevation",
        "sea_surface_elevation_anomaly",
        "sea_surface_height_above_geoid",
    ]
    _unit = ureg.m


class XCurrent(OceanParameter):
    name = "x_current"
    _long_name = "eastward_current_component"
    _standard_name = [
        "sea_water_x_velocity",
        "x_sea_water_velocity",
    ]
    _unit = ureg.m / ureg.s


class YCurrent(OceanParameter):
    name = "y_current"
    _long_name = "northward_current_component"
    _standard_name = [
        "sea_water_y_velocity",
        "y_sea_water_velocity",
    ]
    _unit = ureg.m / ureg.s


class Current(OceanParameter):
    name = "current"
    _long_name = "current_speed"
    _standard_name = "sea_water_speed"
    _unit = ureg.m / ureg.s


class CurrentDir(OceanParameter):
    name = "current_dir"
    _long_name = "current_direction"
    _standard_name = [
        "sea_water_velocity_to_direction",
        "sea_water_to_direction",
        "direction_of_sea_water_velocity",
    ]
    _unit = ureg.deg


class CurrentDirFrom(OceanParameter):
    name = "current_dir"
    _long_name = "current_direction"
    _standard_name = [
        "sea_water_velocity_from_direction",
        "sea_water_from_direction",
    ]
    _unit = ureg.deg


class IceFraction(OceanParameter):
    name = "ice_fraction"
    _long_name = "sea_ice_fraction"
    _standard_name = "sea_ice_area_fraction"
    _unit = ureg.percent


class IceThickness(OceanParameter):
    name = "ice_thickness"
    _long_name = "sea_ice_thickness"
    _standard_name = "sea_ice_thickness"
    _unit = ureg.m



import geo_parameters as gp


def test_no_role():
    assert gp.wave.Hs.i_am() is None
    assert gp.wave.SprPSwell.i_am() is None


def test_period():
    assert gp.wave.Tp().i_am() == "period"
    assert gp.wave.Fp().i_am() == "frequency"
    assert gp.wave.Wp().i_am() == "angular_frequency"


def test_is_x():
    assert gp.wind.XWind.i_am() == "x"
    assert gp.wind.XGust().i_am() == "x"
    assert gp.wave.XStokes().i_am() == "x"
    assert gp.ocean.XCurrent().i_am() == "x"
    assert gp.wind.XFrictionVelocity.i_am() == "x"


def test_is_y():
    assert gp.wind.YWind.i_am() == "y"
    assert gp.wind.YGust().i_am() == "y"
    assert gp.wave.YStokes().i_am() == "y"
    assert gp.ocean.YCurrent().i_am() == "y"
    assert gp.wind.YFrictionVelocity.i_am() == "y"


def test_is_mag():
    assert gp.wind.Wind.i_am() == "magnitude"
    assert gp.wind.Gust().i_am() == "magnitude"
    assert gp.wave.Stokes().i_am() == "magnitude"
    assert gp.ocean.Current().i_am() == "magnitude"
    assert gp.wind.FrictionVelocity.i_am() == "magnitude"


def test_is_dir():
    assert gp.wind.WindDir.i_am() == "direction"
    assert gp.wind.GustDir().i_am() == "direction"
    assert gp.wave.StokesDir().i_am() == "direction"
    assert gp.ocean.CurrentDir().i_am() == "direction"
    assert gp.wind.FrictionVelocityDir.i_am() == "direction"


def test_is_opposite_dir():
    assert gp.wind.WindDirTo.i_am() == "opposite_direction"
    assert gp.wind.GustDirTo().i_am() == "opposite_direction"
    assert gp.wave.StokesDirFrom().i_am() == "opposite_direction"
    assert gp.ocean.CurrentDirFrom().i_am() == "opposite_direction"
    assert gp.wind.FrictionVelocityDirTo.i_am() == "opposite_direction"


def test_get_components_of_self():
    param = gp.wind.XWind
    assert param.my_family().get("x") == param
    assert param.my_family().get("y") == gp.wind.YWind
    assert param.my_family().get("magnitude") == gp.wind.Wind
    assert param.my_family().get("direction") == gp.wind.WindDir
    assert param.my_family().get("opposite_direction") == gp.wind.WindDirTo

    param = gp.wind.XFrictionVelocity
    assert param.my_family().get("x") == param
    assert param.my_family().get("y") == gp.wind.YFrictionVelocity
    assert param.my_family().get("magnitude") == gp.wind.FrictionVelocity
    assert param.my_family().get("direction") == gp.wind.FrictionVelocityDir
    assert param.my_family().get("opposite_direction") == gp.wind.FrictionVelocityDirTo



import geo_parameters as gp


def test_name_of_class():
    assert gp.wave.Hs.name == "hs"
    assert gp.wave.Hs.long_name() == "significant_wave_height"
    assert gp.wave.Hs.standard_name() == "sea_surface_wave_significant_height"
    assert (
        gp.wave.Hs.standard_name(alias=True)
        == "significant_height_of_wind_and_swell_waves"
    )
    assert gp.wave.Efth_max.name == "efth_max"
    assert (
        gp.wave.Efth_max.standard_name()
        == "sea_surface_wave_energy_at_variance_spectral_density_maximum"
    )
    assert gp.wave.Ef_max.name == "emax"
    assert (
        gp.wave.Ef_max.standard_name()
        == "sea_surface_wave_energy_at_variance_spectral_density_maximum"
    )


def test_units_of_class():
    assert str(gp.wave.Hs.units()) == "m"
    assert str(gp.wave.Efth_max.units()) == "m**2*s/rad"
    assert str(gp.wave.Ef_max.units()) == "m**2*s"


def test_name_of_instance():
    assert gp.wave.Hs().name == "hs"
    assert gp.wave.Hs(name="hs").name == "hs"
    assert gp.wave.Hs("hs").name == "hs"
    assert gp.wave.Hs().long_name() == "significant_wave_height"
    assert gp.wave.Hs().standard_name() == "sea_surface_wave_significant_height"
    assert (
        gp.wave.Hs().standard_name(alias=True)
        == "significant_height_of_wind_and_swell_waves"
    )


def test_units_of_instance():
    assert str(gp.wave.Hs().units()) == "m"



import geo_parameters as gp

# import xarray as xr
import numpy as np
import sys


def test_names():
    assert gp.wave.Hs.standard_name() == "sea_surface_wave_significant_height"
    assert gp.wave.Hs.name == "hs"
    hsig = gp.wave.Hs()
    assert hsig.name == "hs"

    hsig = gp.wave.Hs(name="hsig")
    assert hsig.name == "hsig"

    assert hsig.standard_name() == "sea_surface_wave_significant_height"
    assert hsig.long_name() == "significant_wave_height"

    assert (
        hsig.standard_name(alias=True) == "significant_height_of_wind_and_swell_waves"
    )

    assert (
        gp.wave.Tp.standard_name(alias=True)
        == "sea_surface_wave_period_at_variance_spectral_density_maximum"
    )

    assert (
        gp.wave.Tp.standard_name()
        == "sea_surface_wave_period_at_variance_spectral_density_maximum"
    )


def test_get():
    assert gp.get("sea_surface_wave_significant_height") == gp.wave.Hs
    assert gp.get("significant_height_of_wind_and_swell_waves") == gp.wave.Hs
    assert gp.get("hs") is None
    assert gp.shortget("hs") == gp.wave.Hs

    # <class 'geo_parameters.geo_parameters.wave.Hs'>


def test_dicts():
    gp.dict_of_parameters()
    gp.dict_of_parameters(short=True)
    gp.dict_of_parameters(alias=True)


def test_ds_find_me():
    if "xr" in sys.modules:
        ds = xr.Dataset(
            data_vars=dict(
                interesting_hs_name=(["lat", "lon"], np.zeros((11, 6))),
                peak_period=(["lat", "lon"], np.ones((11, 6))),
            ),
            coords=dict(
                lon=(["lon"], np.linspace(5, 10, 6)),
                lat=(["lat"], np.linspace(50, 60, 11)),
            ),
        )
        ds.interesting_hs_name.attrs = {
            "standard_name": "sea_surface_wave_significant_height"
        }

        ds.peak_period.attrs = {
            "standard_name": "sea_surface_wave_period_at_variance_spectral_density_maximum"
        }
        assert gp.wave.Hs.find_me_in_ds(ds) == "interesting_hs_name"
        assert gp.wave.Tp.find_me_in_ds(ds) == "peak_period"



from geo_parameters.metaparameter import MetaParameter
from geo_parameters.ureg import ureg
from typing import Optional, Union
from .relationships import _get_family_dict, _verify_param_type


class WindParameter(MetaParameter):
    @classmethod
    def my_family(
        cls, param_type: Optional[str] = None
    ) -> Union[dict[str, type["WindParameter"]], type["WindParameter"], None]:
        """Returns the dictonary containing the parameters where cls is in.
        Use .my_family('direction') to get the parameter isntead of a dict"""

        _verify_param_type(param_type)
        family_dict = _get_family_dict(cls)

        if param_type is None:  # Return entire family_dict
            return_dict = {}
            for key, value in family_dict.items():
                # E.g. eval("XWind"), which can't be done outside of this module
                return_dict[key] = eval(value)
            return return_dict
        else:  # Retrun class for requested parameter type
            return eval(family_dict.get(param_type, "None"))


class XWind(WindParameter):
    name = "x_wind"
    _long_name = "x_wind_component"
    _standard_name = [
        "x_wind",
        "grid_eastward_wind",
    ]
    _unit = ureg.m / ureg.s


class YWind(WindParameter):
    name = "y_wind"
    _long_name = "y_wind_component"
    _standard_name = [
        "y_wind",
        "grid_northward_wind",
    ]
    _unit = ureg.m / ureg.s


class Wind(WindParameter):
    name = "ff"
    _long_name = "wind_speed"
    _standard_name = "wind_speed"
    _unit = ureg.m / ureg.s


class WindDir(WindParameter):
    name = "dd"
    _long_name = "wind_direction"
    _standard_name = "wind_from_direction"
    _unit = ureg.deg


class WindDirTo(WindParameter):
    name = "dd"
    _long_name = "wind_direction"
    _standard_name = "wind_to_direction"
    _unit = ureg.deg


class XGust(WindParameter):
    name = "x_gust"
    _long_name = "x_gust_component"
    _standard_name = "x_wind_gust"
    _unit = ureg.m / ureg.s


class YGust(WindParameter):
    name = "y_gust"
    _long_name = "y_gust_component"
    _standard_name = "y_wind_gust"
    _unit = ureg.m / ureg.s


class Gust(WindParameter):
    name = "gust"
    _long_name = "wind_gust"
    _standard_name = "wind_speed_of_gust"
    _unit = ureg.m / ureg.s


class GustDir(WindParameter):
    name = "gust_dir"
    _long_name = "wind_gust_direction"
    _standard_name = "wind_gust_from_direction"
    _unit = ureg.deg


class GustDirTo(WindParameter):
    name = "gust_dir"
    _long_name = "wind_gust_direction"
    _standard_name = "wind_gust_to_direction"
    _unit = ureg.deg
    _cf = False


class XFrictionVelocity(WindParameter):
    name = "x_fv"
    _long_name = "x_friction_velocity"
    _standard_name = "eastward_friction_velocity_in_air"
    _unit = ureg.m / ureg.s


class YFrictionVelocity(WindParameter):
    name = "y_fv"
    _long_name = "y_friction_velocity"
    _standard_name = "northward_friction_velocity_in_air"
    _unit = ureg.m / ureg.s


class FrictionVelocity(WindParameter):
    name = "fv"
    _long_name = "friction_velocity"
    _standard_name = "friction_velocity_in_air"
    _unit = ureg.m / ureg.s
    _cf = False


class FrictionVelocityDir(WindParameter):
    name = "fv_dir"
    _long_name = "friction_velocity_direction"
    _standard_name = "friction_velocity_in_air_from_direction"
    _unit = ureg.deg
    _cf = False


class FrictionVelocityDirTo(WindParameter):
    name = "fv_dir"
    _long_name = "friction_velocity_direction"
    _standard_name = "friction_velocity_in_air_to_direction"
    _unit = ureg.deg
    _cf = False



from geo_parameters.metaparameter import MetaParameter
from geo_parameters.ureg import ureg
from typing import Optional, Union
from .relationships import _get_family_dict, _verify_param_type


class GridParameter(MetaParameter):
    @classmethod
    def my_family(
        cls, param_type: Optional[str] = None
    ) -> Union[dict[str, type["GridParameter"]], type["GridParameter"], None]:
        """Returns the dictonary containing the parameters where cls is in.
        Use .my_family('direction') to get the parameter isntead of a dict"""

        _verify_param_type(param_type)
        family_dict = _get_family_dict(cls)

        if param_type is None:  # Return entire family_dict
            return_dict = {}
            for key, value in family_dict.items():
                # E.g. eval("Lon"), which can't be done outside of this module
                return_dict[key] = eval(value)
            return return_dict
        else:  # Retrun class for requested parameter type
            return eval(family_dict.get(param_type, "None"))


class Lon(GridParameter):
    name = "lon"
    _long_name = "longitude"
    _standard_name = "longitude"
    _unit = "degrees_east"


class Lat(GridParameter):
    name = "lat"
    _long_name = "latitude"
    _standard_name = "latitude"
    _unit = "degrees_north"


class X(GridParameter):
    name = "x"
    _long_name = "x_distance"
    _standard_name = "distance_in_x_direction"
    _unit = ureg.m
    _cf = False


class Y(GridParameter):
    name = "y"
    _long_name = "y_distance"
    _standard_name = "distance_in_y_direction"
    _unit = ureg.m
    _cf = False


class Inds(GridParameter):
    name = "inds"
    _long_name = "index_of_points"
    _standard_name = "index_of_geophysical_points"
    _unit = "-"
    _cf = False



import geo_parameters as gp


def test_no_dir_type():
    assert gp.wave.Hs.dir_type() is None
    assert gp.ocean.SeaLevel.dir_type() is None
    param = gp.wave.Hs()
    param2 = gp.ocean.SeaLevel()

    assert param.dir_type() is None
    assert param2.dir_type() is None

    assert gp.wind.FrictionVelocity.dir_type() is None
    assert gp.wind.XFrictionVelocity.dir_type() is None
    assert gp.wind.YFrictionVelocity.dir_type() is None


def test_dir_type_from():
    assert gp.wave.Dirp.dir_type() == "from"
    assert gp.wave.StokesDirFrom.dir_type() == "from"
    assert gp.ocean.CurrentDirFrom.dir_type() == "from"
    assert gp.wind.WindDir.dir_type() == "from"

    param = gp.wave.Dirp()
    param2 = gp.wave.StokesDirFrom()
    param3 = gp.ocean.CurrentDirFrom()
    param4 = gp.wind.WindDir()

    assert param.dir_type() == "from"
    assert param2.dir_type() == "from"
    assert param3.dir_type() == "from"
    assert param4.dir_type() == "from"

    assert gp.wind.FrictionVelocityDir.dir_type() == "from"


def test_dir_type_to():
    assert gp.wave.DirpTo.dir_type() == "to"
    assert gp.wave.StokesDir.dir_type() == "to"
    assert gp.ocean.CurrentDir.dir_type() == "to"
    assert gp.wind.WindDirTo.dir_type() == "to"

    param = gp.wave.DirpTo()
    param2 = gp.wave.StokesDir()
    param3 = gp.ocean.CurrentDir()
    param4 = gp.wind.WindDirTo()

    assert param.dir_type() == "to"
    assert param2.dir_type() == "to"
    assert param3.dir_type() == "to"
    assert param4.dir_type() == "to"
    assert gp.wind.FrictionVelocityDirTo.dir_type() == "to"



"""This module contains functions to help work with geo-parameters, such as decoding and checking"""

from .metaparameter import MetaParameter


def decode(parameter, init: bool = False) -> tuple[str, MetaParameter]:
    """Returns the name of the geo-parameter and the geo-parameter.
    If a string is given, the string and None is returned, otherwise an error is thrown.

    Examples:
    >> gp.decode(gp.wave.Hs('hsig)) -> ('hsig', gp.wave.Hs('hsig'))
    >> gp.decode('hsig') -> ('hsig', None)
    >> gp.decode(gp.wave.Hs) -> ('hs', gp.wave.Hs)
    >> gp.decode(gp.wave.Hs, init=True) -> ('hs', gp.wave.Hs())
    >> gp.decode(<xr.DataArray>) -> *** TypeError: Can only decode types 'MetaParameter' and 'str', not 'DataArray'
    """

    if is_gp(parameter):
        if init and is_gp_class(parameter):
            parameter = parameter()
        return parameter.name, parameter
    elif isinstance(parameter, str):
        return parameter, None
    else:
        raise TypeError(
            f"Can only decode types 'MetaParameter' and 'str', not '{type(parameter).__name__}'"
        )


def is_gp(parameter) -> bool:
    """Checks if the given objects is an instance or class of a geo-parameter"""
    return is_gp_instance(parameter) or is_gp_class(parameter)


def is_gp_instance(parameter) -> bool:
    """Checks if the given object is an instance of a geo-parameter"""
    return isinstance(parameter, MetaParameter)


def is_gp_class(parameter) -> bool:
    """Checks if the given object is a geo-parameter class (i.e. a subclass of MetaParameter"""
    try:
        return issubclass(parameter, MetaParameter)
    except TypeError:
        return False



from .parameter_funcs import (
    list_of_parameters,
    dict_of_parameters,
    create_parameter_dict,
    get,
    shortget,
)

from .decoders import is_gp, is_gp_instance, is_gp_class, decode

from . import wave, wind, ocean, grid
from .ureg import ureg



"""This module keeps track of the relationships between the parameters"""

from typing import Union

PARAM_TYPES = [
    "magnitude",
    "direction",
    "opposite_direction",
    "x",
    "y",
    "period",
    "frequency",
]

RELATIONSHIPS = [
    {
        "x": "XWind",
        "y": "YWind",
        "magnitude": "Wind",
        "direction": "WindDir",
        "opposite_direction": "WindDirTo",
    },
    {
        "x": "XGust",
        "y": "YGust",
        "magnitude": "Gust",
        "direction": "GustDir",
        "opposite_direction": "GustDirTo",
    },
    {
        "x": "XFrictionVelocity",
        "y": "YFrictionVelocity",
        "magnitude": "FrictionVelocity",
        "direction": "FrictionVelocityDir",
        "opposite_direction": "FrictionVelocityDirTo",
    },
    {
        "x": "XCurrent",
        "y": "YCurrent",
        "magnitude": "Current",
        "direction": "CurrentDir",
        "opposite_direction": "CurrentDirFrom",
    },
    {
        "x": "XStokes",
        "y": "YStokes",
        "magnitude": "Stokes",
        "direction": "StokesDir",
        "opposite_direction": "StokesDirFrom",
    },
    {"period": "Tp", "frequency": "Fp", "angular_frequency": "Wp"},
    {"period": "TpSwell", "frequency": "FpSwell"},
    {"period": "TpSwell1", "frequency": "FpSwell1"},
    {"period": "TpSwell2", "frequency": "FpSwell2"},
    {"period": "TpSwell3", "frequency": "FpSwell3"},
    {"period": "Tm01", "frequency": "Fm", "angular_frequency": "Wm"},
    {"period": "TmSwell", "frequency": "FmSwell"},
    {"period": "TmSwell1", "frequency": "FmSwell1"},
    {"period": "TmSwell2", "frequency": "FmSwell2"},
    {"period": "TmSwell3", "frequency": "FmSwell3"},
]


def _verify_param_type(param_type: str) -> None:
    """Verifies that the given parameter type is known"""
    if param_type is not None and param_type not in PARAM_TYPES:
        raise ValueError(f"'param_type' need to be {PARAM_TYPES}, not {param_type}!")


def _get_family_dict(cls) -> dict:
    """Gets a family dictionary of parameters as strings"""
    for rel_dict in RELATIONSHIPS:
        if type(cls()).__name__ in rel_dict.values():
            return rel_dict

    return {}
